from isaac.listeners import ListenerInterface, ListenOptions


class PyListener(ListenerInterface):
    def __init__(
        self,
        whisper_size: str,
        options: ListenOptions = ListenOptions(time_window_s=3),
    ):
        super().__init__(options)

        from listener import Listener

        self.py_listener = Listener(
            time_window=options.time_window_s,
            speech_handler=options.speech_handler,
            on_speech_start=options.on_speech_start,
            whisper_size=whisper_size,
            en_only=options.en_only,
            # show_download=False
        )

    def listen(self):
        self.py_listener.listen()

    def close(self):
        self.py_listener.close()

    def pause(self):
        self.py_listener.pause()

    def resume(self):
        self.py_listener.resume()



import getpass
import importlib
import json
import os
import re
import shutil
import string
import threading
from glob import glob

from yapper import GeminiModel, GroqModel, PiperVoiceGB, PiperVoiceUS
from yapper.constants import piper_voice_quality_map
from yapper.utils import download_piper_model

import isaac.constants as c
import isaac.globals as glb
import isaac.sync as sync
from isaac.listeners.py_listener import ListenOptions, PyListener
from isaac.speakers.piper import PiperSpeaker
from isaac.speakers.utils import mute
from isaac.thinkers import ThinkerConfig
from isaac.thinkers.gemini import GeminiThinker
from isaac.thinkers.groq import GroqThinker
from isaac.types import SettingsInterface
from isaac.utils import (
    get_piper_voice_enum,
    is_wavefile,
    launch_text_editor,
    safe_print,
    select_from,
)

whisper_options = [
    "tiny",
    "tiny.en",
    "base",
    "base.en",
    "small",
    "small.en",
    "medium",
    "medium",
    "turbo",
    "large",
    "auto",
]
whisper_option_details = [
    "~1GB",
    "~1GB, english only",
    "~1GB",
    "~1GB, english only",
    "~2GB",
    "~2GB, english only",
    "~5GB",
    "~5GB, english only",
    "~6GB",
    "~10GB",
    "select based on resource availability",
]


class Settings(SettingsInterface):
    """
    stores the user's preferences, loads settings from the default settings
    file on initialization and can dump back to the same file.
    """

    def __init__(self):
        self.ensure_files()
        cache = json.load(open(c.FILE_SETTINGS, encoding="utf-8"))
        self.groq_key = cache[c.STNG_FLD_GROQ][c.STNG_FLD_KEY]
        self.groq_model = cache[c.STNG_FLD_GROQ][c.STNG_FLD_MODEL]
        self.gemini_key = cache[c.STNG_FLD_GEMINI][c.STNG_FLD_KEY]
        self.gemini_model = cache[c.STNG_FLD_GEMINI][c.STNG_FLD_MODEL]
        self.hearing_enabled = cache[c.STNG_FLD_HEARING][c.STNG_FLD_IS_ENABLED]
        self.whisper_size = cache[c.STNG_FLD_HEARING][c.STNG_FLD_WHISPER_SIZE]
        self.speech_enabled = cache[c.STNG_FLD_SPEECH][c.STNG_FLD_IS_ENABLED]
        self.custom_voices = glob(os.path.join(c.VOICES_DIR, "*.wav"))
        # this field was renamed, this is for backward-compatibility
        self.voice = cache[c.STNG_FLD_SPEECH].get(
            c.STNG_FLD_VOICE, PiperVoiceUS.HFC_FEMALE.value
        )
        self.response_generator = cache[c.STNG_FLD_RSPNS_GENERATOR]
        self.system_message = cache[c.STNG_FLD_SYS_MESSAGE]
        self.context_enabled = cache[c.STNG_FLD_CONTEXT_ENABLED]
        self.shell = cache[c.STNG_FLD_SHELL]
        self.prompt_tokens = 0
        self.completion_tokens = 0
        if self.response_generator is None:
            self.select_lm_provider()

    @property
    def lang_model(self):
        if self.response_generator == c.RSPNS_GNRTR_GEMINI:
            return self.gemini_model
        return self.groq_model

    def ensure_files(self):
        """
        ensures that the settings file exists, if not creates it
        with default values.
        """
        os.makedirs(c.APP_DIR, exist_ok=True)
        os.makedirs(c.VOICES_DIR, exist_ok=True)
        if not os.path.isfile(c.FILE_SETTINGS):
            with open(c.FILE_SETTINGS, "w", encoding="utf-8") as f:
                json.dump(
                    {
                        c.STNG_FLD_GROQ: {
                            c.STNG_FLD_KEY: None,
                            c.STNG_FLD_MODEL: None,
                        },
                        c.STNG_FLD_GEMINI: {
                            c.STNG_FLD_KEY: None,
                            c.STNG_FLD_MODEL: None,
                        },
                        c.STNG_FLD_SPEECH: {
                            c.STNG_FLD_IS_ENABLED: False,
                            c.STNG_FLD_VOICE: PiperVoiceUS.HFC_FEMALE.value,
                        },
                        c.STNG_FLD_HEARING: {
                            c.STNG_FLD_IS_ENABLED: False,
                            c.STNG_FLD_WHISPER_SIZE: "auto",
                        },
                        c.STNG_FLD_RSPNS_GENERATOR: None,
                        c.STNG_FLD_SYS_MESSAGE: None,
                        c.STNG_FLD_CONTEXT_ENABLED: True,
                        c.STNG_FLD_SHELL: c.FILE_SHELL,
                    },
                    f,
                    indent=2,
                )

    def dump_to_cache(self):
        """dumps settings to the default settings file."""
        cache = {}
        cache[c.STNG_FLD_GROQ] = {
            c.STNG_FLD_KEY: self.groq_key,
            c.STNG_FLD_MODEL: self.groq_model,
        }
        cache[c.STNG_FLD_GEMINI] = {
            c.STNG_FLD_KEY: self.gemini_key,
            c.STNG_FLD_MODEL: self.gemini_model,
        }
        cache[c.STNG_FLD_SPEECH] = {
            c.STNG_FLD_IS_ENABLED: self.speech_enabled,
            c.STNG_FLD_VOICE: self.voice,
        }
        cache[c.STNG_FLD_HEARING] = {
            c.STNG_FLD_IS_ENABLED: self.hearing_enabled,
            c.STNG_FLD_WHISPER_SIZE: self.whisper_size,
        }
        cache[c.STNG_FLD_RSPNS_GENERATOR] = self.response_generator
        cache[c.STNG_FLD_SYS_MESSAGE] = self.system_message
        cache[c.STNG_FLD_CONTEXT_ENABLED] = self.context_enabled
        cache[c.STNG_FLD_SHELL] = self.shell
        json.dump(
            cache, open(c.FILE_SETTINGS, "w", encoding="utf-8"), indent=2
        )

    def select_lm_provider(self):
        """
        lets the user select a language model provider,
        currently from `Groq` and `Gemini`.
        """
        options = [c.RSPNS_GNRTR_GEMINI, c.RSPNS_GNRTR_GROQ]
        idx = select_from(
            options, prompt="select an LLM API provider", allow_none=False
        )
        self.response_generator = options[idx]
        self.dump_to_cache()
        if self.lang_model is None:
            self.select_lm()
        self.initialize_thinker()

    def select_lm(self):
        """
        lets the user select a language model to be used for answering
        queries.
        """
        if self.response_generator is None:
            self.select_lm_provider()

        options = (
            [model.value for model in GroqModel]
            if self.response_generator == c.RSPNS_GNRTR_GROQ
            else [model.value for model in GeminiModel]
        )
        idx = select_from(
            options, prompt="please select a language model", allow_none=False
        )
        if self.response_generator == c.RSPNS_GNRTR_GEMINI:
            self.gemini_model = options[idx]
            if self.gemini_key is None:
                self.set_key()
        else:
            self.groq_model = options[idx]
            if self.groq_key is None:
                self.set_key()

        self.initialize_thinker()
        self.dump_to_cache()

    def set_key(self):
        """sets the key for the currently selected language model provider."""
        if self.response_generator is None:
            self.select_lm()
        while True:
            key = getpass.getpass(
                f"please enter your {self.response_generator} key: "
            ).strip()
            if len(key) == 0:
                safe_print("invalid key")
                continue
            if self.response_generator == c.RSPNS_GNRTR_GROQ:
                self.groq_key = key
            else:
                self.gemini_key = key
            break
        self.dump_to_cache()
        safe_print()

    def instruct_lm(self):
        """
        sets the system message to be used for querying the language
        model.
        """
        while True:
            message = launch_text_editor(self.system_message)
            if len(message.strip()) == 0:
                safe_print("empty instruction")
                continue
            break
        self.system_message = message
        self.initialize_thinker()
        self.dump_to_cache()

    def toggle_context(self):
        """
        toggles the use of context to help the model come up with coherent
        responses.
        """
        self.context_enabled = not self.context_enabled
        self.dump_to_cache()

    def enable_speech(self):
        """
        enables speech for the assistant so it can both prints and speaks its
        response.
        """
        with sync.stdout_lock:
            if is_wavefile(self.voice):
                # dynamic import because PocketSpeaker is heavy sl
                from isaac.speakers.pocket import PocketSpeaker

                glb.speaker = PocketSpeaker(self.voice)
            else:
                voice = get_piper_voice_enum(self.voice)
                onnx_f, conf_f = download_piper_model(
                    voice, piper_voice_quality_map[voice], True
                )
                glb.speaker = PiperSpeaker(onnx_f, conf_f)
        self.speech_enabled = True
        self.dump_to_cache()

    def disable_speech(self):
        """disables speech for the assistant."""
        self.speech_enabled = False
        self.dump_to_cache()

    def select_voice(self):
        """
        lets the user select a piper voice for the assistant to speak with
        """
        voices = (
            ["custom voice (clone)"]
            + self.custom_voices
            + [voice.value for voice in PiperVoiceUS]
            + [voice.value for voice in PiperVoiceGB]
        )
        idx = select_from(voices, prompt="select a voice")
        if idx == -1:
            return

        if idx > 0:
            self.voice = voices[idx]
        else:
            if importlib.util.find_spec("pocket_tts") is None:
                safe_print(
                    "please run `pip install py-isaac[custom-voice]` first"
                )
                return
            while True:
                safe_print("custom voice (wave file): ", end="")
                file = os.path.abspath(input().strip())
                if not is_wavefile(file):
                    safe_print("invalid file, please input a valid wave file")
                    continue
                tgt_file = os.path.join(c.VOICES_DIR, os.path.basename(file))
                shutil.copy2(file, tgt_file)
                self.custom_voices.append(tgt_file)
                self.voice = tgt_file
                break

        if not self.speech_enabled:
            self.dump_to_cache()
        else:
            self.enable_speech()

    def toggle_speech(self):
        """toggles the assistant's ability to speak."""
        if self.speech_enabled:
            self.disable_speech()
        else:
            self.enable_speech()

    def select_whisper_size(self):
        """
        lets the user select the size of the whisper model used for converting
        user's speech to text.
        """
        display_options = [
            f"{whisper_options[idx]} ({whisper_option_details[idx]})"
            for idx in range(len(whisper_options))
        ]
        idx = select_from(
            display_options,
            prompt=(
                "please select a whisper model, "
                "large model means better accuracy"
            ),
        )
        self.whisper_size = whisper_options[idx]
        self.dump_to_cache()
        if self.hearing_enabled:
            # load the newly selected whisper model
            self.disable_hearing()
            self.enable_hearing()

    def enable_hearing(self):
        """enables the assistant to hear user with py-listener."""
        event_completion = threading.Event()

        def handle_speech(query: list[str]):
            # print voice transcription on the console so
            # the user may know what the assistant interprets
            query = " ".join(query).strip()
            if re.search(r"\w+", query) is None:
                return
            splits = query.split(" ")
            if len(splits) == 1:
                candidate = "".join(
                    c for c in splits[0] if c not in string.punctuation
                )
                candidate = ":" + candidate.lower()
                if candidate in c.commands:
                    query = candidate
            safe_print(re.sub(r"\n+", ";", query))
            glb.query_queue.put((query, event_completion))
            event_completion.wait()
            if not glb.event_exit.is_set():
                safe_print(">> ", end="")
                event_completion.clear()

        glb.listener = PyListener(
            self.whisper_size,
            ListenOptions(speech_handler=handle_speech, on_speech_start=mute),
        )
        glb.listener.listen()
        self.hearing_enabled = True
        self.dump_to_cache()

    def disable_hearing(self):
        """stops listening to user."""
        glb.listener.close()
        glb.listener = None
        self.hearing_enabled = False
        self.dump_to_cache()

    def toggle_hearing(self):
        """toggles assistant's ability to hear user's speech."""
        if self.hearing_enabled:
            self.disable_hearing()
        else:
            self.enable_hearing()

    def initialize_thinker(self):
        if self.response_generator == c.RSPNS_GNRTR_GEMINI:
            config = ThinkerConfig(
                self.gemini_model, self.gemini_key, self.system_message
            )
            glb.thinker = GeminiThinker(config)
        else:
            config = ThinkerConfig(
                self.groq_model, self.groq_key, self.system_message
            )
            glb.thinker = GroqThinker(config)

    def enact(self):
        """brings the settings into action."""
        if self.hearing_enabled:
            self.enable_hearing()
        self.initialize_thinker()
        if self.speech_enabled:
            self.enable_speech()



import re
from typing import Optional, Union

import isaac.globals as glb


def format_messages(
    query: str,
    fld_role: str,
    fld_content: str,
    role_user: str,
    role_assistant: str,
    role_system: Optional[str] = None,
    system_message: Optional[Union[str, dict]] = None,
    past_exchanges: Optional[list] = [],
) -> list:
    messages = []
    if system_message:
        messages.append({fld_role: role_system, fld_content: system_message})
    for user_msg, assistant_msg in past_exchanges:
        messages.append({fld_role: role_user, fld_content: user_msg})
        messages.append({fld_role: role_assistant, fld_content: assistant_msg})
    messages.append({fld_role: role_user, fld_content: query})
    return messages


def pre_query():
    if glb.settings.lang_model is None:
        glb.settings.select_lm()


def post_query(query: str, answer: str):
    max_answer_words = 30
    max_exchanges = 5
    code_blocks = re.findall(r"```.*?```", answer, re.DOTALL)

    if len(code_blocks) > 0:
        answer = "\n\n".join(code_blocks)
    else:
        parts = answer.split()
        if len(parts) > max_answer_words:
            answer = " ".join(parts[:max_answer_words]) + "..."
    glb.past_exchanges.append((query, answer))
    if len(glb.past_exchanges) > max_exchanges:
        glb.past_exchanges = glb.past_exchanges[-max_exchanges:]


def update_token_cost(prompt: int, completion: int):
    glb.settings.prompt_tokens += prompt
    glb.settings.completion_tokens += completion



try:
    import readline
except ImportError:
    import pyreadline3 as readline
import os
import platform
import shlex
import subprocess
from difflib import SequenceMatcher
from typing import Optional

import psutil

import isaac.constants as c
import isaac.globals as glb
import isaac.sync as sync
from isaac.speakers.utils import mute
from isaac.theme import BOLD_BRIGHT, BRIGHT, RESET
from isaac.thinkers.utils import post_query, pre_query
from isaac.utils import clear, handle_lm_response, label_switch, safe_print


def is_command(text: str) -> bool:
    """
    prints if the given text is a command, i.e. if it starts with a colon(:).
    """
    return text.startswith(":")


def is_shell_command(text: str) -> bool:
    return text.startswith("!")


def command_exists(word: str) -> bool:
    """
    checks if the given word is one of the valid commands defined in the file.
    """
    return word in c.commands


def handle_misspell(word: str) -> str:
    """prints the most similar command to word."""

    # penalizes command matches that don't share the first letter
    def key(c: str):
        return SequenceMatcher(None, word, c).ratio() * (
            0.8 if c[:2] != word[:2] else 1
        )

    similar = max(c.commands, key=key)
    safe_print(f"command not found, did you mean '{similar}'?")


def command_completer(text: str, state: int) -> Optional[str]:
    """handles auto-completion of commands and their arguments."""
    text = readline.get_line_buffer().lstrip()
    if not is_command(text):
        return
    words = shlex.split(text)
    if len(words) > 2:
        return
    command = words[0]
    if not command_exists(command):
        if len(words) == 1:
            options = [cmd for cmd in c.commands if cmd.startswith(command)]
            return options[state] if state < len(options) else None
        return None
    if command not in c.command_args:
        return
    arg = words[1] if len(words) == 2 else ""
    options = [
        option for option in c.command_args[command] if option.startswith(arg)
    ]
    return options[state] if state < len(options) else None


def handle_select(args: list[str]):
    """handles the ':select' command."""
    if len(args) > 1:
        safe_print(":select only takes one argument")
        return
    elif len(args) == 0:
        safe_print(":select needs an argument")
        return

    arg = args[0]
    if arg not in c.selectables:
        safe_print(f"invalid argument {arg}, must be one of {c.selectables}")
        return

    if arg == c.SELECTABLE_LM_PROVIDER:
        glb.settings.select_lm_provider()
    elif arg == c.SELECTABLE_LANG_MODEL:
        glb.settings.select_lm()
    elif arg == c.SELECTABLE_VOICE:
        glb.settings.select_voice()
    else:
        glb.settings.select_whisper_size()


def handle_toggle(args: list[str]):
    """handles the ':toggle' command."""
    if len(args) > 1:
        safe_print(":toggle only takes one argument")
        return
    elif len(args) == 0:
        safe_print(":toggle needs an argument")
        return

    arg = args[0]
    if arg not in c.togglables:
        safe_print(
            f"invalid argument {arg}, argument must be one of {c.togglables}"
        )
        return

    if arg == c.TOGGLABLE_SPEECH:
        glb.settings.toggle_speech()
    elif arg == c.TOGGLABLE_CONTEXT:
        glb.settings.toggle_context()
    else:
        glb.settings.toggle_hearing()


def handle_cmd():
    """handles the `:cmd` command, i.e. launches the appropriate shell."""
    try:
        if glb.settings.hearing_enabled:
            glb.listener.pause()
        if platform.system() == "Windows":
            subprocess.run(["powershell.exe"])
        else:
            subprocess.run(["/bin/sh"])
    finally:
        if glb.settings.hearing_enabled:
            glb.listener.resume()


def display_status():
    """Displays the user's preferences, token cost, and memory consumption."""
    settings = glb.settings
    lang_model = (
        glb.settings.groq_model
        if settings.response_generator == c.RSPNS_GNRTR_GROQ
        else glb.settings.gemini_model
    )
    total_mem = psutil.virtual_memory().total
    parent = psutil.Process(os.getpid())
    consumed_mem = parent.memory_info().rss
    for child in parent.children(recursive=True):
        consumed_mem += child.memory_info().rss
    mem_per = consumed_mem / total_mem * 100

    lines = [
        "%slanguage model:%s" % (BOLD_BRIGHT, RESET),
        "  %sprovider%s: %s" % (BRIGHT, RESET, settings.response_generator),
        "  %smodel:%s %s" % (BRIGHT, RESET, lang_model),
        "  %sinstruction:%s %s"
        % (BRIGHT, RESET, settings.system_message or "null"),
        "  %scontext:%s %s"
        % (BRIGHT, RESET, label_switch(settings.context_enabled)),
        "%sspeech:%s" % (BOLD_BRIGHT, RESET),
        "  %sstatus:%s %s"
        % (BRIGHT, RESET, label_switch(settings.speech_enabled)),
        "  %svoice:%s %s" % (BRIGHT, RESET, glb.settings.voice),
        "%shearing:%s" % (BOLD_BRIGHT, RESET),
        "  %sstatus:%s %s"
        % (BRIGHT, RESET, label_switch(settings.hearing_enabled)),
        "  %smodel:%s %s" % (BRIGHT, RESET, settings.whisper_size or "null"),
        "%sconsumption:%s" % (BOLD_BRIGHT, RESET),
        "  %sprompt tokens:%s %s" % (BRIGHT, RESET, settings.prompt_tokens),
        "  %scompletion tokens:%s %s"
        % (BRIGHT, RESET, settings.completion_tokens),
        "  %smemory:%s %.2f%%" % (BRIGHT, RESET, mem_per),
    ]
    safe_print("\n".join(lines))


def display_commands():
    """Displays the available commands."""
    lines = [
        "%s%s%s to turn features on or off"
        % (BOLD_BRIGHT, c.CMD_TOGGLE, RESET),
        "  %s%s %s%s to toggle the assistant's speech"
        % (BRIGHT, c.CMD_TOGGLE, c.TOGGLABLE_SPEECH, RESET),
        (
            "  %s%s %s%s to toggle the use of conversation"
            " history for coherent responses"
        )
        % (BRIGHT, c.CMD_TOGGLE, c.TOGGLABLE_CONTEXT, RESET),
        "  %s%s %s%s to toggle the assistant's ability to hear you"
        % (BRIGHT, c.CMD_TOGGLE, c.TOGGLABLE_HEARING, RESET),
        "%s%s%s for selecting from available models and voices"
        % (BOLD_BRIGHT, c.CMD_SELECT, RESET),
        "  %s%s %s%s to select the language model provider"
        % (BRIGHT, c.CMD_SELECT, c.SELECTABLE_LM_PROVIDER, RESET),
        "  %s%s %s%s to select the model for generating responses"
        % (BRIGHT, c.CMD_SELECT, c.SELECTABLE_LANG_MODEL, RESET),
        "  %s%s %s%s to select the assistant's voice"
        % (BRIGHT, c.CMD_SELECT, c.SELECTABLE_VOICE, RESET),
        "  %s%s %s%s to select the model interpreting speech"
        % (BRIGHT, c.CMD_SELECT, c.SELECTABLE_WHISPER_MODEL, RESET),
        "%s%s%s to set the LLM API key for the selected provider"
        % (BOLD_BRIGHT, c.CMD_KEY, RESET),
        "%s%s%s to instruct the model to behave a certain way"
        % (BOLD_BRIGHT, c.CMD_INSTRUCT, RESET),
        "%s%s%s to display status and settings"
        % (BOLD_BRIGHT, c.CMD_STATUS, RESET),
        "%s%s%s to mute the assistant" % (BOLD_BRIGHT, c.CMD_MUTE, RESET),
        "%s%s%s to launch a shell session" % (BOLD_BRIGHT, c.CMD_CMD, RESET),
        "%s%s%s to print this help message"
        % (BOLD_BRIGHT, c.CMD_COMMANDS, RESET),
        "%s%s%s to clear the terminal" % (BOLD_BRIGHT, c.CMD_CLEAR, RESET),
        "%s%s%s to exit" % (BOLD_BRIGHT, c.CMD_EXIT, RESET),
    ]
    safe_print("\n".join(lines))


def handle_exit():
    glb.settings.dump_to_cache()
    if glb.settings.hearing_enabled:
        glb.listener.close()
        del glb.listener
    if glb.settings.speech_enabled:
        mute()
        sync.speech_thread and sync.speech_thread.join()
        glb.speaker.close()
        del glb.speaker


def handle_command(words: list[str]):
    """handles words as command."""
    command = words[0]

    if command == c.CMD_SELECT:
        handle_select(words[1:])
    elif command == c.CMD_TOGGLE:
        handle_toggle(words[1:])
    elif command == c.CMD_KEY:
        glb.settings.set_key()
    elif command == c.CMD_INSTRUCT:
        glb.settings.instruct_lm()
    elif command == c.CMD_STATUS:
        display_status()
    elif command == c.CMD_MUTE:
        mute()
    elif command == c.CMD_CMD:
        handle_cmd()
    elif command == c.CMD_COMMANDS:
        display_commands()
    elif command == c.CMD_CLEAR:
        clear()
    elif command == c.CMD_EXIT:
        handle_exit()


def run_query(query: str):
    if len(query.strip()) > 0:
        if is_shell_command(query):
            with sync.stdout_lock:
                subprocess.call(query[1:].lstrip(), shell=True)
        elif is_command(query):
            try:
                words = shlex.split(query)
                cmd_word = words[0]
                if command_exists(cmd_word):
                    handle_command(words)
                else:
                    handle_misspell(cmd_word)
            except ValueError:
                safe_print("invalid command")
        else:
            pre_query()
            try:
                answer = glb.thinker.think(query)
                handle_lm_response(answer)
                post_query(query, answer)
            except Exception as e:
                safe_print(e)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterator, Optional


@dataclass
class ThinkerConfig:
    model: str
    api_key: str
    system_message: str


class ThoughtOptions:
    system_message: Optional[str] = None
    history: Optional[list] = None
    temperature: float = 1.0


class ThinkerInterface(ABC):
    config: ThinkerConfig

    def __init__(self, config: ThinkerConfig):
        self.config = config

    @abstractmethod
    def think(
        self, query: str, options: Optional[ThoughtOptions] = None
    ) -> str: ...

    def stream(self, query: str, options: ThoughtOptions) -> Iterator[str]:
        raise NotImplemented(
            f"{self.__class__.__name__} does not support streaming"
        )



import queue
import threading
from typing import Optional

from isaac.listeners import ListenerInterface
from isaac.speakers import SpeakerInterface
from isaac.thinkers import ThinkerInterface
from isaac.types import SettingsInterface

settings: Optional[SettingsInterface] = None
listener: Optional[ListenerInterface] = None
thinker: Optional[ThinkerInterface] = None
speaker: Optional[SpeakerInterface] = None
query_queue = queue.Queue()
event_exit = threading.Event()
past_exchanges = []



from typing import Optional

import requests

import isaac.constants as c
import isaac.globals as glb
from isaac.thinkers import ThinkerConfig, ThinkerInterface, ThoughtOptions
from isaac.thinkers.utils import update_token_cost


class GeminiThinker(ThinkerInterface):
    def __init__(self, config: ThinkerConfig):
        super().__init__(config)

    def think(
        self, query: str, options: Optional[ThoughtOptions] = None
    ) -> str:
        base = "https://generativelanguage.googleapis.com/v1beta/models"
        url = f"{base}/{self.config.model}:generateContent?key={self.config.api_key}"
        data = {}
        data[c.GMNI_FLD_SYS_INST] = {
            c.GMNI_FLD_PARTS: {
                c.GMNI_FLD_TEXT: (options and options.system_message)
                or self.config.system_message
                or "Hi"
            }
        }
        past_exchanges = (
            ((options and options.history) or glb.past_exchanges)
            if glb.settings.context_enabled
            else []
        )
        contents = []
        for user_msg, assistant_msg in past_exchanges:
            contents.append(
                {
                    c.GMNI_FLD_ROLE: c.GMNI_ROLE_USER,
                    c.GMNI_FLD_PARTS: [{c.GMNI_FLD_TEXT: user_msg}],
                }
            )
            contents.append(
                {
                    c.GMNI_FLD_ROLE: c.GMNI_ROLE_MODEL,
                    c.GMNI_FLD_PARTS: [{c.GMNI_FLD_TEXT: assistant_msg}],
                }
            )
        contents.append(
            {
                c.GMNI_FLD_ROLE: c.GMNI_ROLE_USER,
                c.GMNI_FLD_PARTS: [{c.GMNI_FLD_TEXT: query}],
            }
        )
        data[c.GMNI_FLD_CONTENTS] = contents
        response = requests.post(url, json=data).json()
        if c.GMNI_FLD_ERROR in response:
            return response[c.GMNI_FLD_ERROR].get(
                c.GMNI_FLD_MESSAGE, c.MSG_LANG_MODEL_ERROR
            )
        update_token_cost(
            response[c.GMNI_FLD_USAGE][c.GMNI_USG_PROMPT],
            response[c.GMNI_FLD_USAGE][c.GMNI_USG_COMPLETION],
        )
        content = response[c.GMNI_FLD_CANDIDATES][0][c.GMNI_FLD_CONTENT]
        return content[c.GMNI_FLD_PARTS][0][c.GMNI_FLD_TEXT]



from importlib.util import find_spec
import sys


def main():
    if find_spec("piper") is None:
        print(
            (
                "please ensure piper-tts is installed, "
                "it was not installed by I.S.A.A.C because "
                "of some complications with piper-tts installation.\n"
                'run "pip install piper-phonemize-fix==1.2.1 --no-deps piper-tts==1.2.0".'
            )
        )
        sys.exit(1)

    from isaac.cli import main as cli_main

    cli_main()


if __name__ == "__main__":
    main()



import threading as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional


@dataclass
class SpeechOptions:
    sample_rate: int = 44100
    num_channels: int = 1


class SpeakerInterface(ABC):
    @abstractmethod
    def say(
        self, text: str, options: Optional[SpeechOptions] = None
    ) -> None: ...

    @abstractmethod
    def say_in_thread(
        self, text: str, options: Optional[SpeechOptions] = None
    ) -> t.Thread: ...

    def close(self):
        pass



import queue
import shlex
import threading
import time

try:
    import readline
except ImportError:
    import pyreadline3 as readline

from typing import Optional

import isaac.command as command
import isaac.constants as c
import isaac.globals as glb
from isaac.settings import Settings
from isaac.utils import clear, print_welcome, safe_print


def query_handler():
    while not glb.event_exit.is_set():
        try:
            query, event = glb.query_queue.get(timeout=1)
        except queue.Empty:
            continue
        try:
            if query == c.CMD_EXIT:
                glb.event_exit.set()
                event.set()
            command.run_query(query)
        finally:
            if query != c.CMD_EXIT:
                event.set()


def command_completer(text: str, state: int) -> Optional[str]:
    """handles auto-completion of commands and their arguments."""
    text = readline.get_line_buffer().lstrip()
    if not command.is_command(text):
        return
    words = shlex.split(text)
    if len(words) > 2:
        return
    cmd_word = words[0]
    if not command.command_exists(cmd_word):
        if len(words) == 1:
            options = [cmd for cmd in c.commands if cmd.startswith(cmd_word)]
            return options[state] if state < len(options) else None
        return None
    if cmd_word not in c.command_args:
        return
    arg = words[1] if len(words) == 2 else ""
    options = [
        option for option in c.command_args[cmd_word] if option.startswith(arg)
    ]
    return options[state] if state < len(options) else None


def main():
    """starts the REPL loop."""
    try:
        delims = readline.get_completer_delims().replace(":", "")
        readline.set_completer_delims(delims)
    except AttributeError:
        # might fail on windows
        pass

    readline.set_completer(command_completer)
    readline.parse_and_bind("tab: complete")

    clear()
    print_welcome()
    glb.settings = Settings()
    glb.settings.enact()
    query_thread = threading.Thread(target=query_handler)
    event_completion = threading.Event()
    query_thread.start()

    while True:
        try:
            if glb.settings.hearing_enabled:
                safe_print(">> ", end="")
            while glb.settings.hearing_enabled:
                try:
                    time.sleep(1)
                except KeyboardInterrupt:
                    break
            if glb.event_exit.is_set():
                return
            query = input("" if glb.settings.hearing_enabled else ">> ")
            glb.query_queue.put((query.strip(), event_completion))
            event_completion.wait()
            if glb.event_exit.is_set():
                return
            event_completion.clear()
        except KeyboardInterrupt:
            safe_print("\r", end="")



import os
import platform
import re
import socket
import subprocess
import sys
import tempfile
import wave
from typing import Optional, Union

from rich.console import Console, ConsoleOptions
from rich.markdown import Markdown
from rich.segment import Segment
from yapper import PiperVoiceGB, PiperVoiceUS

import isaac.constants as c
import isaac.globals as glb
import isaac.sync as sync
import isaac.theme as theme

rich_console = Console()


# --- general ---


def clear():
    """
    clears the screen.
    """
    os.system("cls" if platform.system() == "Windows" else "clear")


def safe_print(*args, **kwargs):
    """
    prints the given text on the screen with a lock that prevents
    the listener thread from interrupting it.
    """
    with sync.stdout_lock:
        print(*args, **kwargs, flush=True)


def label_switch(switch: bool):
    """
    return a string that prints the given boolean switch on the screen
    in color.
    """
    label = "on" if switch else "off"
    pre = theme.GREEN if switch else theme.RED
    label = pre + label + theme.RESET
    return label


def print_welcome():
    """Prints the welcome message on the screen."""
    banner = """
%s ___   ____      _         _      ____
|_ _| / ___|    / \       / \    / ___|
 | |  \___ \   / _ \     / _ \  | |
 | | _ ___) | / ___ \ _ / ___ \ | |___
|___(_)____(_)_/   \_(_)_/   \_(_)____|%s"""
    message = banner + "   type %s%s%s to see commands." + "\n"
    message = message % (
        theme.BRIGHT,
        theme.RESET,
        theme.BOLD_BRIGHT,
        c.CMD_COMMANDS,
        theme.RESET,
    )
    print(message)


def safe_input(message: str) -> str:
    try:
        if glb.settings.hearing_enabled:
            glb.listener.pause()
        with sync.stdout_lock:
            return input(message)
    finally:
        if glb.settings.hearing_enabled:
            glb.listener.resume()


def select_from(
    options: list[str],
    prompt: Optional[str] = None,
    allow_none: bool = True,
    none_label: str = "none",
) -> int:
    """
    Displays indexed options on the screen and lets the user select
    one by typing its index.
    """
    if allow_none:
        options += [none_label]
    options = [f"{idx}: {option}" for idx, option in enumerate(options)]
    if prompt is not None:
        safe_print(theme.BOLD_BRIGHT + prompt + theme.RESET)
    safe_print("\n".join(options))
    while True:
        idx = safe_input("select by typing the index: ").strip()
        try:
            idx = int(idx)
            if idx >= len(options):
                raise ValueError()
        except ValueError:
            safe_print("invalid option")
            continue
        if allow_none and idx == len(options) - 1:
            return -1
        safe_print()
        return idx


def check_internet():
    try:
        socket.getaddrinfo("google.com", 80)
        return True
    except OSError:
        return False


def launch_text_editor(initial: Optional[str] = None) -> str:
    with tempfile.NamedTemporaryFile(
        mode="w+", delete=False, suffix=".txt"
    ) as tf:
        if initial is not None:
            tf.write(initial)
        filename = tf.name

    if sys.platform.startswith("win"):
        editor = "notepad"
    elif os.system("which nano > /dev/null 2>&1") == 0:
        editor = "nano"
    else:
        editor = "vi"

    subprocess.call([editor, filename])
    with open(filename, "r") as f:
        content = f.read()

    os.unlink(filename)
    return content


# --- speech ---


def get_piper_voice_enum(str_voice: str) -> Union[PiperVoiceUS, PiperVoiceGB]:
    """
    takes stringified name of a piper voice and return the matching attribute
    from enum `PiperVoiceUS` or `PiperVoiceGB`.
    """
    for voice_enum in list(PiperVoiceUS) + list(PiperVoiceGB):
        if voice_enum.value == str_voice:
            return voice_enum


def is_wavefile(file: str) -> bool:
    try:
        with wave.open(file, "rb"):
            return True
    except Exception:
        return False


def normalize_md(text: str) -> str:
    """
    Detects code blocks in markdown text and replaces the code with
    "you can see the code on the screen", replacs 'code' with specific
    language if it can be detected.
    """

    def replacer(match):
        match = re.search(r"```(\w+).*", text)
        repl = match.group(1) if match else "the code"
        return f"you can see {repl} on the screen. "

    text = re.sub(r"```.*?```", replacer, text, flags=re.DOTALL)
    text = re.sub(r"#+ ", "", text)
    text = re.sub(r"\*\*(.*?)\*\*", r"\1", text)
    text = re.sub(r"\*(.*?)\*", r"\1", text)
    return text


def handle_lm_response(response: str):
    """
    prints the response on the screen, normalizes and speaks it.
    """
    with sync.stdout_lock:
        rich_console.print(CustomMarkdown(response))
    if glb.settings.speech_enabled:
        glb.speaker.say_in_thread(normalize_md(response))


# --- presentation ---


class CustomMarkdown(Markdown):
    """
    rich.markdown.Markdown prepends a leading space character before lines of
    code, that makes the code clearer to read but it becomes difficult to just
    copy it from the console and paste because of the leading space, this
    class detects those leading spaces, and moves them with sequence of spaces
    or before the next newline character, makes code easy to copy and paste.
    """

    def __rich_console__(self, console: Console, options: ConsoleOptions):
        rows: list[list[Segment]] = [[]]
        bg_style = None
        for seg in super().__rich_console__(console, options):
            rows[-1].append(seg)
            if (
                bg_style is None
                and seg.style
                and seg.style.bgcolor
                and seg.text.strip(" ") == ""
            ):
                bg_style = seg.style
            if seg.text.endswith("\n"):
                rows.append([])

        rows = rows[:-1] if len(rows[-1]) == 0 else rows

        for row_idx in range(len(rows)):
            row = rows[row_idx]
            if (
                not row[0].text.startswith(" ")
                or row[0].text.count(" ") == options.max_width
            ):
                continue
            if row[0].text == " ":
                rows[row_idx] = row = row[1:]
            else:
                segment = Segment(
                    row[0].text[1:], row[0].style, row[0].control
                )
                rows[row_idx][0] = row[0] = segment
            rows[row_idx] = (
                row[:-1] + [Segment(" ", style=bg_style)] + [row[-1]]
            )

        for segment in [s for row in rows for s in row]:
            yield segment



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Optional


@dataclass
class ListenOptions:
    sampling_rate: int = 16000
    num_channels: int = 1
    time_window_s: int = 2
    on_speech_start: Optional[Callable] = None
    speech_handler: Optional[Callable[[str], None]] = None
    en_only: bool = True


class ListenerInterface(ABC):
    options: ListenOptions

    def __init__(self, options: ListenOptions):
        self.options = options

    @abstractmethod
    def listen(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def pause(self) -> None: ...

    @abstractmethod
    def resume(self) -> None: ...



import isaac.sync as sync


def mute():
    """
    sets the `event_mute` event that signals the speaker thread to stop
    speaking.
    """
    sync.event_mute.set()


def unmute():
    """
    clears the 'event_mute' event that signals speaker thread to stop.
    """
    sync.event_mute.clear()


def pre_speech():
    if sync.speech_thread and sync.speech_thread.is_alive():
        mute()
        sync.speech_thread.join()
    unmute()



from pathlib import Path

from setuptools import find_packages, setup

ROOT = Path(__file__).parent.resolve()
meta = {}
exec((ROOT / "isaac/meta.py").read_text(), meta)
NAME, VERSION, GITHUB = meta["name"], meta["version"], meta["github"]
long_description = (ROOT / "README.md").read_text()


setup(
    name=NAME,
    version=VERSION,
    author="Nitesh Yadav",
    author_email="nitesh.txt@gmail.com",
    description="Cooles locally run AI assistant",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPL-3.0-or-later",
    url=GITHUB,
    packages=find_packages(),
    install_requires=[
        "sounddevice",
        "py-listener",
        "yapper-tts",
        "psutil",
        "numpy",
        "rich",
        "requests",
        "pyreadline3; sys_platform == 'win32'",
    ],
    extras_require={"custom-voice": ["pocket-tts"]},
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "isaac=isaac.__main__:main",
        ],
    },
    platforms="Posix; Windows",
    keywords=["AI assistant", "local assistant"],
    python_requires=">=3.9",
)



import threading
import wave
from typing import Optional

import sounddevice as sd

import isaac.sync as sync
from isaac.speakers import SpeakerInterface, SpeechOptions
from isaac.speakers.utils import pre_speech


class PocketSpeaker(SpeakerInterface):
    def __init__(
        self,
        voice_file: str,
        options: SpeechOptions = SpeechOptions(sample_rate=24000),
    ):
        from pocket_tts import TTSModel

        self.tts = TTSModel.load_model()
        self.options = options
        device = next(self.tts.parameters()).device
        self.voice_sample = self._load_wav(
            voice_file, self.tts.sample_rate, device
        )

    def _load_wav(self, path: str, sample_rate: int, device):
        import numpy as np
        import torch
        from scipy.signal import resample_poly

        with wave.open(path, "rb") as wf:
            sr = wf.getframerate()
            n_channels = wf.getnchannels()
            n_frames = wf.getnframes()
            audio_bytes = wf.readframes(n_frames)
            audio = (
                np.frombuffer(audio_bytes, dtype=np.int16).astype(np.float32)
                / 32768.0
            )
            if n_channels > 1:
                audio = audio.reshape(-1, n_channels).mean(axis=1)
            if sr != sample_rate:
                gcd = np.gcd(sr, sample_rate)
                up = sample_rate // gcd
                down = sr // gcd
                audio = resample_poly(audio, up, down)

        return torch.from_numpy(audio).unsqueeze(0).to(device)

    def say(self, text: str, options: Optional[SpeechOptions] = None) -> None:
        with sd.OutputStream(
            samplerate=self.tts.sample_rate, channels=1, dtype="float32"
        ) as stream:
            state = self.tts.get_state_for_audio_prompt(self.voice_sample)
            with sync.speech_lock:
                for chunk in self.tts.generate_audio_stream(
                    model_state=state, text_to_generate=text, copy_state=True
                ):
                    if sync.event_mute.is_set():
                        break
                    chunk = (
                        chunk.numpy()
                        if chunk.device.type == "cpu"
                        else chunk.cpu().numpy()
                    )
                    stream.write(chunk)

    def say_in_thread(
        self, text: str, options: Optional[SpeechOptions] = None
    ) -> threading.Thread:
        pre_speech()
        t = threading.Thread(target=self.say, args=(text, options))
        sync.speech_thread = t
        t.start()
        return t

    def close(self):
        del self.tts



from abc import ABC
from typing import Literal, Optional


class SettingsInterface(ABC):
    groq_key: Optional[str]
    groq_model: str
    gemini_key: Optional[str]
    gemini_model: str
    hearing_enabled: bool
    whisper_size: str
    speech_enabled: bool
    custom_voices: list[str]
    piper_voice: str
    response_generator: Literal["gemini", "groq"]
    system_message = Optional[str]
    context_enabled: bool
    shell: str
    prompt_tokens: int
    completion_tokens: int



import threading
from typing import Optional

import numpy as np
import sounddevice as sd
from piper.voice import PiperVoice

import isaac.sync as sync
from isaac.speakers import SpeakerInterface, SpeechOptions
from isaac.speakers.utils import pre_speech


class PiperSpeaker(SpeakerInterface):
    def __init__(
        self,
        onnx_f: str,
        conf_f: str,
        options: SpeechOptions = SpeechOptions(
            sample_rate=22050, num_channels=1
        ),
    ):
        self.speaker = PiperVoice.load(onnx_f, conf_f)
        self.options = options

    def say(self, text: str, options: Optional[SpeechOptions] = None) -> None:
        """Speaks the given text in a separate thread."""
        options = options or self.options
        with sd.OutputStream(
            samplerate=options.sample_rate,
            channels=options.num_channels,
            dtype="int16",
        ) as stream:
            with sync.speech_lock:
                for audio_bytes in self.speaker.synthesize_stream_raw(text):
                    if sync.event_mute.is_set():
                        break
                    stream.write(np.frombuffer(audio_bytes, dtype=np.int16))

    def say_in_thread(
        self, text: str, options: Optional[SpeechOptions] = None
    ) -> threading.Thread:
        pre_speech()
        t = threading.Thread(target=self.say, args=(text, options))
        sync.speech_thread = t
        t.start()
        return t



import os
import os.path as path
import sys

import isaac.meta as meta

# Gemini API fields
GMNI_FLD_SYS_INST = "system_instruction"
GMNI_FLD_CONTENT = "content"
GMNI_FLD_ROLE = "role"
GMNI_FLD_PARTS = "parts"
GMNI_FLD_CONTENTS = "contents"
GMNI_FLD_CANDIDATES = "candidates"
GMNI_FLD_TEXT = "text"
GMNI_FLD_USAGE = "usageMetadata"
GMNI_FLD_ERROR = "error"
GMNI_FLD_MESSAGE = "message"

GMNI_USG_PROMPT = "promptTokenCount"
GMNI_USG_COMPLETION = "candidatesTokenCount"

GMNI_ROLE_USER = "user"
GMNI_ROLE_MODEL = "model"

# Groq  API fields
GROQ_FLD_ROLE = "role"
GROQ_FLD_CHOICES = "choices"
GROQ_FLD_MESSAGE = "message"
GROQ_FLD_CONTENT = "content"
GROQ_FLD_MESSAGES = "messages"
GROQ_FLD_MODEL = "model"
GROQ_FLD_USAGE = "usage"
GROQ_FLD_ERROR = "error"
GROQ_FLD_ERROR_CODE = "code"

GROQ_USG_PROMPT = "prompt_tokens"
GROQ_USG_COMPLETION = "completion_tokens"

GROQ_ROLE_SYSTEM = "system"
GROQ_ROLE_USER = "user"
GROQ_ROLE_ASSISTANT = "assistant"

# Settings fields
STNG_FLD_GROQ = "groq"
STNG_FLD_GEMINI = "gemini"
STNG_FLD_KEY = "key"
STNG_FLD_MODEL = "model"
STNG_FLD_HEARING = "hearing"
STNG_FLD_WHISPER_SIZE = "whisperSize"
STNG_FLD_SPEECH = "speech"
STNG_FLD_VOICE = "voice"
STNG_FLD_IS_ENABLED = "isEnabled"
STNG_FLD_RSPNS_GENERATOR = "responseGenerator"
STNG_FLD_SYS_MESSAGE = "systemMessage"
STNG_FLD_CONTEXT_ENABLED = "contextEnabled"
STNG_FLD_SHELL = "shell"

# Response generator APIs
RSPNS_GNRTR_GROQ = "groq"
RSPNS_GNRTR_GEMINI = "gemini"


if os.name == "nt":
    APP_DIR = path.join(os.getenv("APPDATA"), meta.name)
elif os.name == "posix":
    home = path.expanduser("~")
    if os.uname().sysname == "Darwin":
        APP_DIR = path.join(home, "Library/Application Support", meta.name)
    else:
        APP_DIR = path.join(home, ".config", meta.name)
else:
    print("your system is not supported")
    sys.exit()

VOICES_DIR = os.path.join(APP_DIR, "voices")

# File names
FILE_SETTINGS = path.join(APP_DIR, "settings.json")
FILE_SHELL = "powershell.exe" if os.name == "nt" else "/bin/sh"

# messages
MSG_LANG_MODEL_ERROR = "could not process that query, something went wrong"
MSG_NO_INTERNET = (
    "you seem to be offline, please check your internet connection"
)

# commands
CMD_SELECT = ":select"
CMD_TOGGLE = ":toggle"
CMD_KEY = ":key"
CMD_INSTRUCT = ":instruct"
CMD_STATUS = ":status"
CMD_MUTE = ":mute"
CMD_CMD = ":cmd"
CMD_COMMANDS = ":commands"
CMD_CLEAR = ":clear"
CMD_EXIT = ":exit"

SELECTABLE_LM_PROVIDER = "lm_provider"
SELECTABLE_VOICE = "voice"
SELECTABLE_LANG_MODEL = "lm"
SELECTABLE_WHISPER_MODEL = "whisper"

TOGGLABLE_SPEECH = "speech"
TOGGLABLE_HEARING = "hearing"
TOGGLABLE_CONTEXT = "context"

commands = {
    CMD_SELECT,
    CMD_TOGGLE,
    CMD_KEY,
    CMD_INSTRUCT,
    CMD_STATUS,
    CMD_MUTE,
    CMD_CMD,
    CMD_COMMANDS,
    CMD_CLEAR,
    CMD_EXIT,
}
selectables = [
    SELECTABLE_LM_PROVIDER,
    SELECTABLE_LANG_MODEL,
    SELECTABLE_VOICE,
    SELECTABLE_WHISPER_MODEL,
]
togglables = [TOGGLABLE_SPEECH, TOGGLABLE_HEARING, TOGGLABLE_CONTEXT]
command_args = {CMD_SELECT: selectables, CMD_TOGGLE: togglables}



from typing import Optional

import requests

import isaac.constants as c
import isaac.globals as glb
from isaac.thinkers import ThinkerConfig, ThinkerInterface, ThoughtOptions
from isaac.thinkers.utils import format_messages, update_token_cost


class GroqThinker(ThinkerInterface):
    def __init__(self, config: ThinkerConfig):
        super().__init__(config)

    def think(
        self, query: str, options: Optional[ThoughtOptions] = None
    ) -> str:
        url = "https://api.groq.com/openai/v1/chat/completions"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.config.api_key}",
            "User-Agent": "curl/7.68.0",
        }
        past_exchanges = (
            ((options and options.history) or glb.past_exchanges)
            if glb.settings.context_enabled
            else []
        )
        messages = format_messages(
            query=query,
            fld_role=c.GROQ_FLD_ROLE,
            fld_content=c.GROQ_FLD_CONTENT,
            role_user=c.GROQ_ROLE_USER,
            role_assistant=c.GROQ_ROLE_ASSISTANT,
            role_system=c.GROQ_ROLE_SYSTEM,
            system_message=(options and options.system_message)
            or self.config.system_message,
            past_exchanges=past_exchanges,
        )
        data = {
            c.GROQ_FLD_MESSAGES: messages,
            c.GROQ_FLD_MODEL: self.config.model,
        }
        response = requests.post(url, json=data, headers=headers).json()
        if c.GROQ_FLD_ERROR in response:
            if (
                response[c.GROQ_FLD_ERROR].get(c.GROQ_FLD_ERROR_CODE)
                == "model_decommissioned"
            ):
                return (
                    f"the model '{self.config.model}' has been decommissioned"
                    ", please use another Groq model."
                )
            return c.MSG_LANG_MODEL_ERROR
        update_token_cost(
            response[c.GROQ_FLD_USAGE][c.GROQ_USG_PROMPT],
            response[c.GROQ_FLD_USAGE][c.GROQ_USG_COMPLETION],
        )
        return response[c.GROQ_FLD_CHOICES][0][c.GROQ_FLD_MESSAGE][
            c.GROQ_FLD_CONTENT
        ]
